# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Confirm task."""

from django.core.exceptions import ValidationError

from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.tasks.wait import Confirm
from debusine.server.tasks.wait.models import ConfirmData, DenyAction
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.models import Confirmation, OutputData
from debusine.test.django import TestCase


class ConfirmTaskTests(TestCase):
    """Test the Confirm task."""

    scenario = scenarios.DefaultScopeUser()

    def test_no_additional_properties(self) -> None:
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(ValidationError, error_msg):
            self.playground.create_wait_task(
                task_name="confirm",
                task_data={"input": {}},
                workflow_data=WorkRequestWorkflowData(needs_input=True),
            )

    def test_label(self) -> None:
        task = self.playground.create_wait_task(
            task_name="confirm",
            task_data={},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        ).get_task()
        self.assertEqual(task.get_label(), "wait for confirmation")

    def test_respond_confirm(self) -> None:
        work_request = self.playground.create_wait_task(
            task_name="confirm",
            workflow_data=WorkRequestWorkflowData(needs_input=True),
            mark_running=True,
        )
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=True)

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertEqual(work_request.completed_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(confirmation=Confirmation(confirmed=True)),
        )

    def test_respond_deny_fail(self) -> None:
        work_request = self.playground.create_wait_task(
            task_name="confirm",
            task_data=ConfirmData(deny_action=DenyAction.FAIL),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
            mark_running=True,
        )
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=False, comment="Broken")

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.FAILURE)
        self.assertEqual(work_request.completed_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )

    def test_respond_deny_abort(self) -> None:
        self.scenario.set_current()
        workflow = self.playground.create_workflow(mark_running=True)
        work_request = workflow.create_child_wait(
            task_name="confirm",
            task_data=ConfirmData(deny_action=DenyAction.ABORT),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(work_request, mark_running=True)
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=False, comment="Broken")

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertEqual(work_request.aborted_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.RUNNING)

    def test_respond_deny_abort_workflow(self) -> None:
        self.scenario.set_current()
        workflow = self.playground.create_workflow(mark_running=True)
        work_request = workflow.create_child_wait(
            task_name="confirm",
            task_data=ConfirmData(deny_action=DenyAction.ABORT_WORKFLOW),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(work_request, mark_running=True)
        task = work_request.get_task()
        assert isinstance(task, Confirm)

        task.respond(user=self.scenario.user, confirm=False, comment="Broken")

        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertEqual(work_request.aborted_by, self.scenario.user)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                confirmation=Confirmation(confirmed=False, comment="Broken")
            ),
        )
        workflow.refresh_from_db()
        self.assertEqual(workflow.status, WorkRequest.Statuses.ABORTED)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing-side no-operation task."""

from pathlib import Path
from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import SigningNoopData
from debusine.tasks import DefaultDynamicData
from debusine.tasks.models import BaseDynamicTaskData


class SigningNoop(
    BaseSigningTask[SigningNoopData, BaseDynamicTaskData],
    DefaultDynamicData[SigningNoopData],
):
    """Task that runs on server-side Celery workers and returns a boolean."""

    TASK_NAME = "noop"
    TASK_VERSION = 1

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
        """Do nothing."""
        return self.data.result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the blhc workflow."""

import datetime as dt
from collections.abc import Sequence
from typing import Any, ClassVar, NewType

from django.test import override_settings
from django.utils import timezone

import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianBlhc,
    DebianSourcePackage,
    TaskTypes,
)
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    TaskDatabase,
    WorkRequest,
    Workspace,
)
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.collections.lookup import LookupResult
from debusine.server.scheduler import schedule
from debusine.server.workflows import BlhcWorkflow, WorkflowValidationError
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    BlhcWorkflowData,
    BlhcWorkflowDynamicData,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestBase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
    BlhcFlags,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry

Architecture = NewType("Architecture", str)


class BlhcWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestBase[BlhcWorkflow]
):
    """Unit tests for :py:class:`BlhcWorkflow`."""

    package_build_log_artifacts: ClassVar[dict[Architecture, Artifact]]
    source_artifact: ClassVar[Artifact]

    workspace: ClassVar[Workspace]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common data."""
        super().setUpTestData()

        cls.workspace = cls.playground.get_default_workspace()

        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )

        cls.package_build_log_artifacts = {}
        for arch in ["amd64", "arm64", "all"]:
            artifact = cls.playground.create_build_log_artifact(
                source="hello", version="1.0-1", build_arch=arch
            )

            cls.package_build_log_artifacts[Architecture(arch)] = artifact

    def orchestrate(
        self,
        *,
        task_data: BlhcWorkflowData,
        source_artifact: Artifact,
        architectures: Sequence[str] | None = None,
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate a BlhcWorkflow."""
        if architectures is None:
            architectures = ["amd64", "arm64", "all"]

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow."""

            TASK_NAME = pipeline_task_name

            def populate(self_) -> None:
                """Populate the pipeline."""
                source_data = source_artifact.create_data()
                assert isinstance(source_data, DebianSourcePackage)
                assert architectures is not None

                sbuild = self_.work_request_ensure_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:trixie",
                        architectures=list(architectures),
                    ),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="sbuild", step="sbuild"
                    ),
                )
                self.playground.advance_work_request(sbuild, mark_running=True)
                for arch in architectures:
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            build_architecture=arch,
                            environment="debian/match:codename=trixie",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.PACKAGE_BUILD_LOG,
                        name=f"buildlog-{arch}",
                        data={
                            "source_package_name": source_data.name,
                            "source_package_version": source_data.version,
                            "architecture": arch,
                        },
                    )
                sbuild.unblock_workflow_children()

                blhc = self_.work_request_ensure_child_workflow(
                    task_name="blhc",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="blhc",
                        step="blhc",
                    ),
                )
                self_.orchestrate_child(blhc)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(orchestrate_workflow(root))

        return root

    def create_blhc_workflow(
        self,
        *,
        binary_artifacts: LookupMultiple | None = None,
        package_build_logs: LookupMultiple | None = None,
        extra_task_data: dict[str, Any] | None = None,
        include_source_artifact: bool = True,
    ) -> BlhcWorkflow:
        """
        Create and schedule a Blhc workflow.

        :param binary_artifacts: binary artifacts for the workflow.
        :param package_build_logs: build logs for the workflow. If both
          ``binary_artifacts`` and ``package_build_logs`` are None, use all
          the build log artifacts created by ``setUpTestData()``.
        :param extra_task_data: any extra task data to pass to the workflow.
        :param include_source_artifact: True to include self.source_artifact.
        """
        if binary_artifacts is None and package_build_logs is None:
            package_build_logs = LookupMultiple(
                tuple(
                    package_build_log.id
                    for package_build_log in (
                        self.package_build_log_artifacts.values()
                    )
                )
            )

        task_data = BlhcWorkflowData(
            source_artifact=(
                self.source_artifact.id if include_source_artifact else None
            ),
            binary_artifacts=binary_artifacts,
            package_build_logs=package_build_logs,
            vendor="debian",
            codename="bookworm",
        )

        if extra_task_data is not None:
            task_data = task_data.model_copy(update=extra_task_data)

        example_pipeline = self.orchestrate(
            task_data=task_data,
            source_artifact=self.source_artifact,
        )
        work_request = example_pipeline.children.get(
            task_name="blhc", task_type=TaskTypes.WORKFLOW
        )
        return self.get_workflow(work_request)

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        *,
        timestamp: dt.datetime | None = None,
    ) -> CollectionItem:
        """Add a blhc result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="blhc", result=WorkRequest.Results.SUCCESS, validate=False
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source=package, version=version, architecture=architecture
                ),
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int((timestamp or timezone.now()).timestamp()),
                "work_request_id": work_request.id,
            },
        )

    @preserve_db_task_registry()
    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        # validate_input() is called and does not raise
        self.create_blhc_workflow()

    @preserve_db_task_registry()
    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        w = self.create_blhc_workflow(
            extra_task_data={"qa_suite": "nonexistent@debian:suite"}
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            w.validate_input()

    @preserve_db_task_registry()
    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        w = self.create_blhc_workflow(
            extra_task_data={
                "reference_qa_results": "nonexistent@debian:qa-results"
            }
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            w.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "other", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT, artifact=build_log
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-2"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT, artifact=build_log
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_unexpected(self) -> None:
        """_has_current_reference_qa_result: unexpected result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        with self.assertRaisesRegex(ValueError, r"^Unexpected result"):
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    artifact=self.playground.create_artifact()[0],
                ),
                "amd64",
            )

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT, artifact=build_log
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_current_promise(self) -> None:
        """_has_current_reference_qa_result: current result matching promise."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        build_log_promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.PACKAGE_BUILD_LOG,
                "source_package_name": "hello",
                "source_package_version": "1.0-1",
            },
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple.model_validate(
                    [f"test@{CollectionCategory.TEST}/name:test"]
                ),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.BARE,
                    collection_item=build_log_promise,
                ),
                "amd64",
            )
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        blhc_workflow = self.create_blhc_workflow(
            extra_task_data={
                "extra_flags": [BlhcFlags.BINDNOW, BlhcFlags.LINE_NUMBERS]
            }
        )

        # One work request per architecture
        self.assertQuerySetEqual(
            blhc_workflow.work_request.children.order_by("id").values_list(
                "task_type", "task_name", "task_data", "workflow_data_json"
            ),
            [
                (
                    TaskTypes.WORKER,
                    "blhc",
                    {
                        "environment": "debian/match:codename=bookworm",
                        "build_architecture": (
                            "amd64" if architecture == "all" else architecture
                        ),
                        "input": {"artifact": f"{artifact.id}@artifacts"},
                        "extra_flags": [
                            BlhcFlags.BINDNOW,
                            BlhcFlags.LINE_NUMBERS,
                        ],
                    },
                    {
                        "display_name": (
                            f"build log hardening check for {architecture}"
                        ),
                        "step": f"blhc-{architecture}",
                    },
                )
                for architecture, artifact in sorted(
                    self.package_build_log_artifacts.items()
                )
            ],
        )

    def test_populate_with_uploads(self) -> None:
        source_upload = self.playground.create_upload_artifacts(
            binary=False
        ).upload
        binary_upload = self.playground.create_upload_artifacts(
            source=False, binaries=[("hello", "amd64")]
        ).upload
        log = self.package_build_log_artifacts[Architecture("amd64")]
        self.playground.create_artifact_relation(
            artifact=log, target=binary_upload
        )
        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                source_artifact=source_upload.id,
                binary_artifacts=LookupMultiple((binary_upload.id,)),
                vendor="debian",
                codename="sid",
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(
            child.task_data["input"], {"artifact": f"{log.id}@artifacts"}
        )

    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), [])

    def test_populate_previous_reference_qa_result_outdated(self) -> None:
        """The workflow produces a reference QA result if it is outdated."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        old_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-2"
        )
        source_item = sid_suite.manager.add_artifact(
            self.playground.create_source_artifact(
                name="hello", version="1.0-2"
            ),
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        [child] = wr.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "blhc")
        self.assertEqual(
            child.task_data,
            {
                "environment": "debian/match:codename=sid",
                "build_architecture": "amd64",
                "input": {"artifact": f"{build_log.id}@artifacts"},
                "extra_flags": [],
            },
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-2",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
            artifact_filters={"category": ArtifactCategory.BLHC},
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[
                ActionSkipIfLookupResultChanged(
                    lookup=(
                        f"sid@{CollectionCategory.QA_RESULTS}/"
                        f"latest:blhc_hello_amd64"
                    ),
                    collection_item_id=old_qa_result.id,
                    promise_name="reference-qa-result|blhc-amd64",
                )
            ],
            on_failure=[qa_result_action],
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    collection="internal@collections",
                    name_template=("reference-qa-result|blhc-amd64"),
                    artifact_filters={"category": ArtifactCategory.BLHC},
                ),
                qa_result_action,
            ],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.BLHC,
            data=DebianBlhc(
                source="hello", version="1.0-2", architecture="amd64"
            ),
            work_request=child,
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/latest:blhc_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.playground.create_debian_environment(
            codename="sid", variant="blhc"
        )
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        racing_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        [child] = wr.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'sid@{CollectionCategory.QA_RESULTS}/"
                    f"latest:blhc_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/latest:blhc_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = ["amd64", "arm64", "i386", "ppc64el", "s390x"]
        reference_source_artifact = self.playground.create_source_artifact(
            name="hello", version="0.9-1"
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                source_artifact=reference_source_artifact.id,
                package_build_logs=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:buildlog-{arch}"
                        for arch in architectures
                        if arch != "ppc64el"
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=reference_source_artifact,
            architectures=architectures,
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=BlhcWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                source_artifact=self.source_artifact.id,
                package_build_logs=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:buildlog-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=self.source_artifact,
            architectures=architectures,
            parent=root,
        )

        # Unblock the blhc tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS
            )

        reference_blhc_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_blhc_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="blhc"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source="hello", version="0.9-1", architecture=architecture
                ),
                workspace=reference.workspace,
                work_request=reference_tasks[architecture],
            )[0]
            for architecture in architectures
            if architecture != "ppc64el"
        }
        new_blhc_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_blhc_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="blhc"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source="hello", version="1.0-1", architecture=architecture
                ),
                workspace=new.workspace,
                work_request=new_tasks[architecture],
            )[0]
            for architecture in architectures
        }

        self.assertIsNone(new_blhc_workflow.output_data)
        self.assertEqual(schedule(), [])
        self.playground.advance_work_request(
            reference_tasks["amd64"], result=WorkRequest.Results.SUCCESS
        )
        self.assertEqual(schedule(), [])
        self.playground.advance_work_request(
            new_tasks["amd64"], result=WorkRequest.Results.SUCCESS
        )

        self.schedule_and_run_workflow_callback(new_blhc_workflow)

        expected_analysis: list[tuple[str, RegressionAnalysisStatus, None]] = [
            ("amd64", RegressionAnalysisStatus.STABLE, None)
        ]
        new_blhc_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_blhc_workflow,
            "0.9-1",
            reference_results,
            "1.0-1",
            new_results,
            expected_analysis,
        )

        expected_callback_statuses = {
            "amd64": WorkRequest.Statuses.COMPLETED,
            "arm64": WorkRequest.Statuses.BLOCKED,
            "i386": WorkRequest.Statuses.BLOCKED,
            "ppc64el": WorkRequest.Statuses.BLOCKED,
            "s390x": WorkRequest.Statuses.BLOCKED,
        }
        self.assert_callback_statuses(
            new_blhc_workflow, expected_callback_statuses
        )

        for architecture, reference_result, new_result, status in (
            (
                "arm64",
                WorkRequest.Results.SUCCESS,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
            ),
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.STABLE,
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture], result=reference_result
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result
            )
            expected_analysis.append((architecture, status, None))
            expected_callback_statuses[architecture] = (
                WorkRequest.Statuses.PENDING
            )

        self.assert_callback_statuses(
            new_blhc_workflow, expected_callback_statuses
        )

        # Running any one of the pending callbacks is enough, but in fact
        # one is scheduled for each architecture and the scheduler will
        # catch up with them over multiple runs.  Simulate this and check
        # that the result is correct at each stage.
        for architecture in ("arm64", "i386", "ppc64el", "s390x"):
            self.schedule_and_run_workflow_callback(new_blhc_workflow)

            new_blhc_workflow.refresh_from_db()
            assert new_blhc_workflow.output_data is not None
            self.assert_regression_analysis(
                new_blhc_workflow,
                "0.9-1",
                reference_results,
                "1.0-1",
                new_results,
                expected_analysis,
            )
            expected_callback_statuses[architecture] = (
                WorkRequest.Statuses.COMPLETED
            )
            self.assert_callback_statuses(
                new_blhc_workflow, expected_callback_statuses
            )

    @preserve_db_task_registry()
    def test_compute_system_required_tags(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        workflow = self.create_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )
        self.assertCountEqual(
            workflow.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    @preserve_db_task_registry()
    def test_compute_dynamic_data_with_binary_artifacts(self) -> None:
        binary_uploads = {
            architecture: self.playground.create_upload_artifacts(
                source=False, binaries=[("hello", architecture)]
            ).upload
            for architecture in ("amd64", "s390x")
        }
        package_build_logs = {
            architecture: self.playground.create_build_log_artifact(
                build_arch=architecture
            )
            for architecture in binary_uploads
        }
        for architecture in binary_uploads:
            self.playground.create_artifact_relation(
                artifact=package_build_logs[architecture],
                target=binary_uploads[architecture],
            )

        workflow = self.create_blhc_workflow(
            binary_artifacts=LookupMultiple(
                tuple(
                    f"{artifact.id}@artifacts"
                    for artifact in binary_uploads.values()
                )
            )
        )

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(workflow.work_request)),
            BlhcWorkflowDynamicData(
                subject="hello",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[
                    artifact.id for artifact in binary_uploads.values()
                ],
                package_build_logs_ids=sorted(
                    [artifact.id for artifact in package_build_logs.values()]
                ),
            ),
        )

    @preserve_db_task_registry()
    def test_compute_dynamic_data_with_package_build_logs(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        workflow = self.create_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(workflow.work_request)),
            BlhcWorkflowDynamicData(
                subject="hello hello-traditional",
                source_artifact_id=self.source_artifact.id,
                package_build_logs_ids=[
                    package_build_log_1.id,
                    package_build_log_2.id,
                ],
            ),
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_with_binary_artifacts(self) -> None:
        binary_uploads = {
            architecture: self.playground.create_upload_artifacts(
                source=False, binaries=[("hello", architecture)]
            ).upload
            for architecture in ("amd64", "s390x")
        }

        workflow = self.create_blhc_workflow(
            binary_artifacts=LookupMultiple(
                tuple(
                    f"{artifact.id}@artifacts"
                    for artifact in binary_uploads.values()
                )
            )
        )

        assert workflow.data.source_artifact is not None
        assert workflow.data.binary_artifacts is not None
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[
                        artifact.id for artifact in binary_uploads.values()
                    ],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_with_package_build_logs(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        workflow = self.create_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )

        assert workflow.data.source_artifact is not None
        assert workflow.data.package_build_logs is not None
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.package_build_logs,
                    label="package_build_logs",
                    artifact_ids=[
                        package_build_log_1.id,
                        package_build_log_2.id,
                    ],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_without_source_artifact(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        workflow = self.create_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            ),
            include_source_artifact=False,
        )

        assert workflow.data.package_build_logs is not None
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple(
                    lookup=workflow.data.package_build_logs,
                    label="package_build_logs",
                    artifact_ids=[
                        package_build_log_1.id,
                        package_build_log_2.id,
                    ],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Server-side database interaction for worker tasks.

Worker tasks mostly run on external workers, but they also run some
pre-dispatch code on the server which needs to access the database.  To
avoid having to import Django from :py:mod:`debusine.tasks`, we define an
interface that is implemented by server code.
"""

from abc import ABC, abstractmethod
from collections.abc import Collection as AbcCollection
from dataclasses import dataclass
from typing import Any, TYPE_CHECKING, overload

from debusine.artifacts.models import (
    ArtifactCategory,
    ArtifactData,
    CollectionCategory,
)
from debusine.client.models import RelationType
from debusine.tasks.models import (
    BackendType,
    LookupMultiple,
    LookupSingle,
    build_key_value_lookup_segment,
    build_lookup_string_segments,
    parse_key_value_lookup_segment,
    parse_lookup_string_segments,
)

if TYPE_CHECKING:
    from debusine.tasks import BaseTask
    from debusine.tasks.executors import ExecutorImageCategory


@dataclass
class ArtifactInfo:
    """Information about an artifact."""

    id: int
    category: str
    data: ArtifactData


class MultipleArtifactInfo(tuple[ArtifactInfo, ...]):
    """Information about multiple artifacts."""

    def get_ids(self) -> list[int]:
        """Return the ID of each artifact."""
        return [item.id for item in self]


@dataclass
class CollectionInfo:
    """Information about a collection."""

    id: int
    scope_name: str
    workspace_name: str
    category: str
    name: str
    data: dict[str, Any]


class TaskDatabaseInterface(ABC):
    """Interface for interacting with the database from worker tasks."""

    @overload
    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo: ...

    @overload
    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: None,
        default_category: CollectionCategory | None = None,
    ) -> None: ...

    @abstractmethod
    def lookup_single_artifact(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> ArtifactInfo | None:
        """
        Look up a single artifact.

        :param lookup: A :ref:`lookup-single`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about the artifact, or None if the provided
          lookup is None (for convenience in some call sites).
        :raises KeyError: if the lookup does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way, or does
          not resolve to an artifact.
        """

    @abstractmethod
    def lookup_multiple_artifacts(
        self,
        lookup: LookupMultiple | None,
        default_category: CollectionCategory | None = None,
    ) -> MultipleArtifactInfo:
        """
        Look up multiple artifacts.

        :param lookup: A :ref:`lookup-multiple`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about each artifact.
        :raises KeyError: if any of the lookups does not resolve to an item.
        :raises LookupError: if any of the lookups is invalid in some way,
          or does not resolve to an artifact.
        """

    @abstractmethod
    def find_related_artifacts(
        self,
        artifact_ids: AbcCollection[int],
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
    ) -> MultipleArtifactInfo:
        """
        Find artifacts via relations.

        For each of the artifacts in ``artifact_ids``, find all artifacts of
        ``target_category`` that are related to the initial artifacts with a
        relation of type ``relation_type``.

        :param artifact_ids: The IDs of artifacts to start from.
        :param target_category: The category of artifacts to find.
        :param relation_type: The type of relation to follow.
        :return: Information about each artifact.
        """

    @overload
    @abstractmethod
    def lookup_single_collection(
        self,
        lookup: LookupSingle,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo: ...

    @overload
    @abstractmethod
    def lookup_single_collection(
        self,
        lookup: None,
        default_category: CollectionCategory | None = None,
    ) -> None: ...

    @abstractmethod
    def lookup_single_collection(
        self,
        lookup: LookupSingle | None,
        default_category: CollectionCategory | None = None,
    ) -> CollectionInfo | None:
        """
        Look up a single collection.

        :param lookup: A :ref:`lookup-single`.
        :param default_category: If the first segment of a string lookup
          (which normally identifies a collection) does not specify a
          category, use this as the default category.
        :return: Information about the collection, or None if the provided
          lookup is None (for convenience in some call sites).
        :raises KeyError: if the lookup does not resolve to an item.
        :raises LookupError: if the lookup is invalid in some way, or does
          not resolve to a collection.
        """

    def lookup_singleton_collection(
        self, category: CollectionCategory
    ) -> CollectionInfo | None:
        """Look up a singleton collection for `category`, if it exists."""
        try:
            return self.lookup_single_collection(f"_@{category}")
        except KeyError:
            return None

    @abstractmethod
    def get_server_setting(self, setting: str) -> str:
        """Look up a Django setting (strings only)."""

    @abstractmethod
    def export_suite_signing_keys(self, suite_id: int) -> str | None:
        """Export signing keys for a suite as an ASCII-armored keyring."""

    def resolve_inputs(self, task: "BaseTask[Any, Any]") -> None:
        """
        Resolve inputs for this task.

        This will look up information required by the input using this task
        database, and make it available to task methods.
        """
        from debusine.tasks.inputs import TaskInput

        TaskInput.resolve_inputs(task, self)

    def get_environment(
        self,
        lookup: LookupSingle,
        *,
        architecture: str | None = None,
        backend: BackendType | None = None,
        default_category: CollectionCategory | None = None,
        image_category: "ExecutorImageCategory | None" = None,
        try_variant: str | None = None,
    ) -> ArtifactInfo:
        """
        Get an environment.

        This automatically fills in some additional constraints if needed.

        :param lookup: the base lookup provided by the task data
        :param architecture: the task's build architecture, if available
        :param backend: the task's backend, or None if the environment lookup
          does not need to be constrained to a particular backend
        :param default_category: the default category to use for the first
          segment of the lookup
        :param image_category: try to use an environment with this image
          category; defaults to the image category needed by the executor for
          `self.backend`
        :param try_variant: None if the environment lookup does not need to be
          constrained to a particular variant; otherwise, and if `lookup` does
          not already specify a variant, then try looking up an environment with
          this variant first, and fall back to looking for an environment with
          no variant
        :return: the ArtifactInfo of a suitable environment artifact
        """
        # Prevent import loop
        from debusine.tasks.executors import (
            ExecutorImageCategory,
            executor_class,
        )

        lookups: list[LookupSingle] = []

        if (
            isinstance(lookup, str)
            and len(segments := parse_lookup_string_segments(lookup)) == 2
            and (parsed := parse_key_value_lookup_segment(segments[1]))[0]
            == "match"
        ):
            # Supplement the environment lookup with the task architecture,
            # if required.
            lookup_type, filters = parsed
            assert lookup_type == "match"

            # TODO: If filters already have values for architecture/format, then
            # check their consistency.

            if architecture is not None:
                filters.setdefault("architecture", architecture)

            if image_category is None and backend is not None:
                image_category = executor_class(backend).image_category
            match image_category:
                case ExecutorImageCategory.TARBALL:
                    filters.setdefault("format", "tarball")
                case ExecutorImageCategory.IMAGE:
                    filters.setdefault("format", "image")
                case _ as unreachable:
                    raise AssertionError(
                        f"Unexpected image category: {unreachable}"
                    )

            if backend is not None:
                filters.setdefault("backend", backend)

            if try_variant is not None and "variant" not in filters:
                for variant in (try_variant, ""):
                    lookups.append(
                        build_lookup_string_segments(
                            segments[0],
                            build_key_value_lookup_segment(
                                lookup_type, {**filters, "variant": variant}
                            ),
                        )
                    )
            else:
                lookups.append(
                    build_lookup_string_segments(
                        segments[0],
                        build_key_value_lookup_segment(lookup_type, filters),
                    )
                )
        else:
            lookups.append(lookup)

        for try_lookup in lookups[:-1]:
            try:
                return self.lookup_single_artifact(
                    try_lookup, default_category=default_category
                )
            except KeyError:
                pass

        return self.lookup_single_artifact(
            lookups[-1], default_category=default_category
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
View extension to model sidebar elements.

Sidebar elements are informational elements or actions that can be used to
populate the UI sidebar.
"""

import datetime as dt
from dataclasses import dataclass
from functools import cached_property

from django.db.models import Model
from django.template.context import BaseContext
from django.utils.formats import date_format
from django.utils.html import format_html
from django.utils.safestring import SafeString
from django.utils.timesince import timesince

from debusine.db.models import (
    Artifact,
    Collection,
    User,
    WorkRequest,
    Worker,
    Workspace,
)
from debusine.web.icons import Icons
from debusine.web.views.base import Widget


@dataclass(kw_only=True, frozen=True)
class SidebarItem(Widget):
    """Renderable element for the right sidebar."""

    #: icon to use
    icon: str = "app"
    #: label describing what the field is about
    label: str
    #: user-presented value for the field
    value: str | None = None
    #: detailed information about the value
    detail: str | None = None

    @cached_property
    def content(self) -> str:
        """Select the main content to display."""
        if self.value:
            return self.value
        if self.detail:
            return self.detail
        return "-"

    @cached_property
    def main_tooltip(self) -> str:
        """Select what to use as tooltip for the whole entry."""
        if self.detail and self.value and self.detail != self.value:
            return self.detail
        return self.label

    @cached_property
    def icon_tooltip(self) -> str | None:
        """Select if a tooltip is shown on the icon, and what it is."""
        if self.main_tooltip == self.label:
            return None
        return self.label

    def render(self, context: BaseContext) -> str:
        """Render this item."""
        raise NotImplementedError(
            f"{self.__class__.__name__}.render not implemented"
        )


@dataclass(kw_only=True, frozen=True)
class SidebarElement(SidebarItem):
    """Informative element for the right sidebar."""

    def render(self, context: BaseContext) -> str:  # noqa: ARG002, U100
        """Render this item."""
        if self.icon_tooltip:
            icon_title = format_html(' title="{}"', self.icon_tooltip)
        else:
            icon_title = SafeString()
        return format_html(
            '<span class="list-group-item" title="{main_tooltip}">'
            '<span class="bi bi-{icon}"{icon_title}></span> '
            '{content}</span>',
            main_tooltip=self.main_tooltip,
            icon_title=icon_title,
            icon=self.icon,
            content=self.content,
        )


@dataclass(kw_only=True, frozen=True)
class SidebarAction(SidebarItem):
    """Action element for the right sidebar."""

    # Action URL
    url: str

    def render(self, context: BaseContext) -> str:  # noqa: ARG002, U100
        """Render this item."""
        if self.icon_tooltip:
            icon_title = format_html(' title="{}"', self.icon_tooltip)
        else:
            icon_title = SafeString()
        return format_html(
            '<a class="list-group-item list-group-item-action'
            ' d-flex justify-content-between"'
            ' title="{main_tooltip}" href="{url}">'
            '<span><span class="bi bi-{icon}"{icon_title}></span> '
            '{content}</span>'
            '<span>…</span>'
            '</a>',
            main_tooltip=self.main_tooltip,
            icon_title=icon_title,
            url=self.url,
            icon=self.icon,
            content=self.content,
        )


def create_workspace(workspace: Workspace) -> SidebarItem:
    """Create workspace information."""
    return SidebarAction(
        url=workspace.get_absolute_url(),
        icon="person-workspace",
        label="Workspace",
        value=workspace.name,
    )


def create_work_request(work_request: WorkRequest) -> SidebarItem:
    """Create work request information."""
    kwargs = {
        "icon": Icons.WORK_REQUEST,
        "label": "Work request",
    }

    kwargs["detail"] = f"{work_request.task_type} {work_request.task_name} task"
    kwargs["value"] = work_request.get_label()

    return SidebarAction(
        url=work_request.get_absolute_url(),
        **kwargs,
    )


def create_user(
    user: User | None, *, context: Model | None = None
) -> SidebarItem:
    """Create user information."""
    match context:
        case WorkRequest():
            title = "User who created the work request"
        case Artifact():
            title = "User who created the artifact"
        case _:
            title = "User"

    return SidebarElement(
        icon=Icons.USER,
        label=title,
        value=str(user) if user else None,
    )


def create_created_at(at: dt.datetime) -> SidebarItem:
    """Create created_at information."""
    return SidebarElement(
        icon=Icons.CREATED_AT,
        label="Created",
        value=timesince(at) + " ago",
        detail=date_format(at, "DATETIME_FORMAT"),
    )


def create_expire_at(at: dt.datetime | None) -> SidebarItem:
    """Create expire_at information."""
    kwargs = {
        "icon": Icons.EXPIRE_AT,
        "label": "Expiration date",
    }
    if at:
        return SidebarElement(
            detail=date_format(at, "DATETIME_FORMAT"),
            value=timesince(at, reversed=True),
            **kwargs,
        )
    else:
        return SidebarElement(value="Never", **kwargs)


def create_work_request_status(work_request: WorkRequest) -> SidebarItem:
    """Create information about a work request status and result."""
    kwargs = {
        "icon": Icons.WORK_REQUEST_STATUS,
        "label": "Status and result",
    }

    match work_request.status:
        case WorkRequest.Statuses.PENDING:
            status_bg = "secondary"
        case WorkRequest.Statuses.RUNNING:
            status_bg = "info"
        case WorkRequest.Statuses.COMPLETED:
            status_bg = "primary"
        case WorkRequest.Statuses.ABORTED | WorkRequest.Statuses.BLOCKED:
            status_bg = "dark"
        case _ as unreachable:  # noqa: F841
            status_bg = "info"

    status = format_html(
        '<span class="badge text-bg-{bg}">{status}</span>',
        bg=status_bg,
        status=work_request.status.capitalize(),
    )

    result_bg: str | None
    match work_request.result:
        case WorkRequest.Results.NONE:
            result_bg = None
        case WorkRequest.Results.SUCCESS:
            result_bg = "success"
        case WorkRequest.Results.FAILURE:
            result_bg = "warning"
        case WorkRequest.Results.ERROR:
            result_bg = "danger"
        case WorkRequest.Results.SKIPPED:
            result_bg = "secondary"
        case _ as unreachable:  # noqa: F841
            result_bg = "info"

    if result_bg is None:
        kwargs["value"] = status
    else:
        kwargs["value"] = status + format_html(
            ' <span class="badge text-bg-{bg}">{result}</span>',
            bg=result_bg,
            result=work_request.result.capitalize(),
        )

    return SidebarElement(**kwargs)


def create_work_request_superseded(work_request: WorkRequest) -> SidebarItem:
    """Create a link to the work request that supersedes this one."""
    kwargs = {
        "icon": Icons.WORK_REQUEST_SUPERSEDED,
        "label": "Superseded by",
    }
    if (superseded := getattr(work_request, "superseded", None)) is None:
        return SidebarElement(**kwargs)
    return SidebarAction(
        url=superseded.get_absolute_url(),
        value=str(superseded),
        **kwargs,
    )


def create_work_request_supersedes(work_request: WorkRequest) -> SidebarItem:
    """Create a link to the work request superseded by this one."""
    kwargs = {
        "icon": Icons.WORK_REQUEST_SUPERSEDES,
        "label": "Supersedes",
    }
    if (supersedes := getattr(work_request, "supersedes", None)) is None:
        return SidebarElement(**kwargs)

    return SidebarAction(
        url=supersedes.get_absolute_url(),
        value=str(supersedes),
        **kwargs,
    )


def create_workflow(workflow: WorkRequest | None) -> SidebarItem:
    """Create a link to the associated workflow."""
    if workflow is None:
        return SidebarElement(
            icon=Icons.WORKFLOW,
            label="Workflow",
        )

    return SidebarAction(
        icon="diagram-3",
        label="Workflow",
        url=workflow.get_absolute_url(),
        value=str(workflow),
    )


def create_worker(worker: Worker | None) -> SidebarItem:
    """Create a link to a worker."""
    kwargs = {
        "icon": Icons.WORKER,
        "label": "Worker",
    }
    if worker is None:
        return SidebarElement(**kwargs)
    return SidebarAction(
        value=worker.name, url=worker.get_absolute_url(), **kwargs
    )


def create_work_request_started_at(work_request: WorkRequest) -> SidebarItem:
    """Show the time the work request spent in a queue before being started."""
    kwargs = {
        "icon": Icons.STARTED_AT,
        "label": "Time from creation to start",
    }
    if not work_request.started_at:
        return SidebarElement(**kwargs)

    return SidebarElement(
        detail=date_format(work_request.started_at, "DATETIME_FORMAT"),
        value=timesince(work_request.created_at, work_request.started_at),
        **kwargs,
    )


def create_work_request_duration(work_request: WorkRequest) -> SidebarItem:
    """Show the duration of the work request run."""
    kwargs = {
        "icon": Icons.DURATION,
        "label": "Duration",
    }

    if (
        work_request.started_at
        and work_request.status == WorkRequest.Statuses.RUNNING
    ):
        return SidebarElement(
            value=timesince(work_request.started_at),
            **kwargs,
        )

    if not work_request.started_at or not work_request.completed_at:
        return SidebarElement(**kwargs)

    return SidebarElement(
        value=timesince(work_request.started_at, work_request.completed_at),
        detail=date_format(work_request.completed_at, "DATETIME_FORMAT"),
        **kwargs,
    )


def create_collection(collection: Collection) -> SidebarItem:
    """Link to a collection."""
    return SidebarAction(
        url=collection.get_absolute_url(),
        icon=Icons.COLLECTION,
        label="Collection",
        detail=f"{collection.name}@{collection.category}",
        value=collection.name,
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for Autopkgtest work request."""

import itertools
import re
from enum import StrEnum
from typing import Any, Optional, TypeAlias, assert_never, override

import pydantic
from more_itertools import peekable

from debusine.artifacts import AutopkgtestArtifact
from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.db.models import Artifact, FileInArtifact, WorkRequest
from debusine.db.models.tasks import DBTask
from debusine.tasks.models import AutopkgtestData, AutopkgtestDynamicData
from debusine.web.views import view_utils
from debusine.web.views.artifacts import ArtifactPlugin
from debusine.web.views.files import LogFileWidget
from debusine.web.views.work_request import Tab, WorkRequestPlugin

DBAutopkgtest: TypeAlias = DBTask[AutopkgtestData, AutopkgtestDynamicData]


class AutopkgtestTab(Tab):
    """Autopkgtest specialized tab."""

    def __init__(self, work_request: WorkRequest, task: DBAutopkgtest) -> None:
        """Define this tab."""
        super().__init__(name="autopkgtest", label="Autopkgtest")
        self.work_request = work_request
        self.task = task

    @staticmethod
    def _get_fail_on_scenarios(fail_on: dict[str, bool]) -> list[str]:
        """Process fail_on task data into a list of strings for viewing."""
        return [
            key.replace("_test", "")
            for key, value in fail_on.items()
            if key.endswith("_test") and value
        ]

    @staticmethod
    def _list_files(artifact: Artifact | None) -> list[str]:
        if artifact is not None:
            return [
                file.path
                for file in artifact.fileinartifact_set.filter(
                    complete=True
                ).order_by("path")
            ]
        else:
            return []

    @property
    def _source_artifact(self) -> dict[str, Any]:
        task_data = self.task.data
        source_artifact: dict[str, Any] = {
            "lookup": task_data.input.source_artifact,
            "id": None,
            "artifact": None,
            "files": [],
        }

        if (dynamic_data := self.task.dynamic_data) is not None:
            pk = dynamic_data.input_source_artifact_id
            source_artifact["id"] = pk
            try:
                source_artifact["artifact"] = artifact = Artifact.objects.get(
                    id=pk
                )
            except Artifact.DoesNotExist:
                pass
            else:
                source_artifact["files"] = self._list_files(artifact)
        return source_artifact

    def _get_request_data(self) -> dict[str, Any]:
        """Return data about the request."""
        task = self.task
        task_data = task.data
        dynamic_data = task.dynamic_data

        # TODO: It would be useful to show the request data after default
        # values have been set, but unfortunately these are only set by the
        # task after it loads data from the database and aren't saved
        # anywhere else.
        request_data: dict[str, Any] = {
            "source_artifact": self._source_artifact,
            "source_name": None,
            "source_version": None,
            "binary_artifacts": {
                "lookup": task_data.input.binary_artifacts.export(),
                "artifacts": [],
            },
            "context_artifacts": {
                "lookup": task_data.input.context_artifacts.export(),
                "artifacts": [],
            },
            "build_architecture": task_data.build_architecture,
            "vendor": None,
            "codename": None,
            "backend": task_data.backend,
            "include_tests": task_data.include_tests,
            "exclude_tests": task_data.exclude_tests,
            "debug_level": task_data.debug_level,
            "extra_repositories": task_data.extra_repositories,
            "use_packages_from_base_repository": (
                task_data.use_packages_from_base_repository
            ),
            "extra_environment": task_data.extra_environment,
            "needs_internet": task_data.needs_internet,
            "fail_on_scenarios": self._get_fail_on_scenarios(
                task_data.fail_on.model_dump()
            ),
            "timeout": task_data.timeout,
        }

        # This information is in the result artifact's data if it exists,
        # but it won't necessarily exist - for example, a testbed failure
        # will leave us without a result artifact.
        if dynamic_data is not None:
            try:
                environment_artifact = Artifact.objects.get(
                    id=dynamic_data.environment_id
                )
            except Artifact.DoesNotExist:
                pass
            else:
                if environment_artifact.category in {
                    ArtifactCategory.SYSTEM_TARBALL,
                    ArtifactCategory.SYSTEM_IMAGE,
                }:
                    request_data["vendor"] = environment_artifact.data.get(
                        "vendor"
                    )
                    request_data["codename"] = environment_artifact.data.get(
                        "codename"
                    )

            if source_artifact := request_data["source_artifact"]["artifact"]:
                if source_artifact.category == ArtifactCategory.SOURCE_PACKAGE:
                    request_data.update(
                        {
                            "source_name": source_artifact.data.get("name"),
                            "source_version": source_artifact.data.get(
                                "version"
                            ),
                        }
                    )

            binary_artifacts = {
                artifact.id: artifact
                for artifact in Artifact.objects.filter(
                    id__in=dynamic_data.input_binary_artifacts_ids
                )
            }
            for binary_artifact_id in dynamic_data.input_binary_artifacts_ids:
                binary_artifact = binary_artifacts.get(binary_artifact_id)
                request_data["binary_artifacts"]["artifacts"].append(
                    {
                        "id": binary_artifact_id,
                        "files": self._list_files(binary_artifact),
                        "artifact": binary_artifact,
                    }
                )

            context_artifacts = {
                artifact.id: artifact
                for artifact in Artifact.objects.filter(
                    id__in=dynamic_data.input_context_artifacts_ids
                )
            }
            for context_artifact_id in dynamic_data.input_context_artifacts_ids:
                context_artifact = context_artifacts.get(context_artifact_id)
                request_data["context_artifacts"]["artifacts"].append(
                    {
                        "id": context_artifact_id,
                        "files": self._list_files(context_artifact),
                    }
                )

        request_data["task_description"] = (
            f"{request_data['source_name'] or 'UNKNOWN'}_"
            f"{request_data['source_version'] or 'UNKNOWN'}_"
            f"{request_data['build_architecture']}"
        )
        if request_data["vendor"] and request_data["codename"]:
            request_data["task_description"] += (
                f" in {request_data['vendor']}:{request_data['codename']}"
            )

        return request_data

    @staticmethod
    def _get_result_data(
        artifact: Artifact,
    ) -> list[tuple[str, str, str | None]]:
        """Return a list of tuples describing individual test results."""
        result_data = []
        for name, result in sorted(artifact.data.get("results", {}).items()):
            result_data.append((name, result["status"], result.get("details")))
        return result_data

    @override
    def get_context_data(self) -> dict[str, Any]:
        result_artifact = (
            self.work_request.artifact_set.filter(
                category=AutopkgtestArtifact._category
            )
            .order_by("id")
            .first()
        )

        context_data: dict[str, Any] = {
            "request_data": self._get_request_data(),
            "result": self.work_request.result,
        }
        if result_artifact:
            context_data["result_artifact"] = result_artifact
            context_data["result_data"] = self._get_result_data(result_artifact)

        return context_data


class AutopkgtestViewWorkRequestPlugin(WorkRequestPlugin):
    """View for Autopkgtest work request."""

    task_type = TaskTypes.WORKER
    task_name = "autopkgtest"
    task: DBAutopkgtest

    @override
    def add_tabs(self, tabs: list[Tab]) -> None:
        tabs.insert(0, AutopkgtestTab(self.work_request, self.task))

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {}


class AutopkgtestLogParser:
    """Parse Autopkgtest logs."""

    class Result(StrEnum):
        PASS = "PASS"
        FAIL = "FAIL"
        SKIP = "SKIP"

    class Section(pydantic.BaseModel):
        """Represents a logical section."""

        text: str
        color: str | None = (
            None  # bootstrap color: "primary", "secondary", "success"...
        )
        subsections: list["AutopkgtestLogParser.SubSection"] = []

    class SubSection(Section):
        """Represents a logical subsection with the test result."""

        line: int  # line number, 1-indexed
        result: Optional["AutopkgtestLogParser.Result"] = None
        expanded: bool = False  # subsection's contents is expanded or not
        include_in_table_of_contents: bool = True
        tooltip: str | None = None

    _PREFIX = r"\s*\d+s\s+autopkgtest\s+\[\d\d:\d\d:\d\d\]:"
    _RE_STARTING = re.compile(rf"{_PREFIX}\s+starting date and time")
    _RE_APT_SOURCE = re.compile(rf"{_PREFIX}\s+@{{20}}\s+apt-source.*")
    _RE_TESTBED_SETUP = re.compile(rf"{_PREFIX}\s+@{{20}} test bed setup")
    _RE_TEST_PREPARING_TESTBED = re.compile(
        rf"{_PREFIX}\s+test\s+(.*):\s+preparing testbed"
    )
    _RE_TEST_STARTS = re.compile(rf"{_PREFIX}\s+test\s+(.*): ")
    _RE_TEST_RUN = re.compile(
        rf"{_PREFIX}\s+test\s+(.*): \[-----------------------"
    )
    _RE_SUMMARY = re.compile(rf"{_PREFIX}\s+@{{20}}\s+(summary)")
    _RE_RESULT_LINE = re.compile(r"^\s*\d+s\s+\S+\s*(PASS|FAIL|SKIP)")

    def __init__(self, log: bytes) -> None:
        """Initialize parser."""
        # splitlines() cannot be used: it's needed to separate by
        # "\n" only and splitlines() also separates by "\r" (e.g. the
        # autopkgtest logs contain a line with "\r" in the output of apt-get
        # The rstrip(b"\n") is to avoid an extra line at the end of the log
        # (splitlines() have the "keepends" parameter for this, but not split())
        self._text = log.decode(errors="replace").rstrip("\n")

    def _parse_preparation(self, lines: "peekable[tuple[int, str]]") -> Section:
        preparation = self.Section(text="Preparation", color="primary")

        for line_number, line in lines:
            # One of the two ways that the "prepare section" finishes is that
            # next line starts with a test
            (_, next_line) = lines.peek((None, ""))

            if self._RE_STARTING.match(line):
                preparation.subsections.append(
                    self.SubSection(
                        text="start run", line=line_number, expanded=True
                    )
                )
            elif self._RE_TESTBED_SETUP.match(line):
                preparation.subsections.append(
                    self.SubSection(text="test bed setup", line=line_number)
                )
            elif self._RE_APT_SOURCE.match(line):
                preparation.subsections.append(
                    self.SubSection(text="apt-source", line=line_number)
                )
                # there will not be more subsections for preparation
                break
            elif self._RE_TEST_STARTS.match(next_line):
                # preparation has finished
                break

        return preparation

    def _parse_tests(self, lines: "peekable[tuple[int, str]]") -> Section:
        tests = self.Section(text="Tests", color="primary")

        current_test_name = None

        for line_number, line in lines:
            if m := self._RE_TEST_PREPARING_TESTBED.match(line):
                current_test_name = m.group(1)
                tests.subsections.append(
                    self.SubSection(
                        text=f"preparing testbed {current_test_name}",
                        line=line_number,
                        include_in_table_of_contents=False,
                    )
                )
            elif (
                self._RE_TEST_RUN.match(line) and current_test_name is not None
            ):
                tests.subsections.append(
                    self.SubSection(text=current_test_name, line=line_number)
                )
            elif (
                m := self._RE_RESULT_LINE.match(line)
            ) and current_test_name is not None:
                subsection = tests.subsections[-1]
                assert isinstance(subsection, self.SubSection)
                Result = self.Result
                result = Result(m.group(1))

                match result:
                    case Result.PASS:
                        subsection.color = "success"
                        subsection.tooltip = "Pass"
                    case Result.SKIP:
                        subsection.color = "secondary"
                        subsection.tooltip = "Skip"
                    case Result.FAIL:
                        subsection.color = "danger"
                        subsection.tooltip = "Fail"
                        subsection.expanded = True
                    case _ as unreachable:
                        assert_never(unreachable)

            # The only way to know that we finished with the tests
            # is to check if next line is the "Summary"
            (_, line) = lines.peek((None, ""))
            if self._RE_SUMMARY.match(line):
                break

        return tests

    def _parse_closing(self, lines: "peekable[tuple[int, str]]") -> Section:
        closing = self.Section(text="Closing", color="primary")

        for line_number, line in lines:
            if self._RE_SUMMARY.match(line):
                closing.subsections.append(
                    self.SubSection(
                        text="summary", line=line_number, expanded=True
                    )
                )

        return closing

    def parse(self) -> tuple[list[Section], int]:
        """Parse log file. Return (list[Section], total_lines)."""
        lines = self._text.split("\n")
        total_lines = len(lines)
        lines_peekable = peekable(enumerate(lines, start=1))

        preparation = self._parse_preparation(lines_peekable)
        tests = self._parse_tests(lines_peekable)
        closing = self._parse_closing(lines_peekable)

        return [preparation, tests, closing], total_lines


class AutopkgtestArtifactPlugin(ArtifactPlugin):
    """Plugin for AutopkgtestArtifact."""

    artifact_category = ArtifactCategory.AUTOPKGTEST
    template_name = "web/artifact-detail.html"
    name = "autopkgtest"

    object_name = "artifact"

    def get_context_data(self) -> dict[str, Any]:
        """Return the context."""
        slug = "autopkgtest-log"

        specialized_tab = {
            "specialized_tab": {
                "label": "Autopkgtest log",
                "slug": slug,
                "template": "web/_autopkgtest-artifact-detail.html",
            }
        }

        try:
            log_file_in_artifact = self.artifact.fileinartifact_set.get(
                path="log"
            )
        except FileInArtifact.DoesNotExist:
            return {
                **specialized_tab,
                "autopkgtest_artifact_error": "Artifact does not contain log",
            }

        try:
            autopkgtest_contents = view_utils.read_file(
                log_file_in_artifact, max_file_size=50 * 1024 * 1024
            )
        except view_utils.FileInArtifactError as exc:
            return {
                **specialized_tab,
                "autopkgtest_artifact_error": exc.to_html(),
            }

        sections, total_lines = AutopkgtestLogParser(
            autopkgtest_contents
        ).parse()

        file_sections = []

        subsections = list(
            itertools.chain.from_iterable(
                log_section.subsections for log_section in sections
            )
        )

        for subsection, subsection_next in itertools.pairwise(subsections):
            file_sections.append(
                LogFileWidget.Section(
                    title=subsection.text,
                    start_line=subsection.line,
                    end_line=subsection_next.line - 1,
                    expanded=subsection.expanded,
                )
            )

        # Add the latest remaining section
        if subsections:
            last_subsection = subsections[-1]
            file_sections.append(
                LogFileWidget.Section(
                    title=last_subsection.text,
                    start_line=last_subsection.line,
                    end_line=total_lines,
                    expanded=last_subsection.expanded,
                )
            )

        log_widget = LogFileWidget(
            log_file_in_artifact,
            sections=file_sections,
            file_contents=autopkgtest_contents,
        )

        return {
            **specialized_tab,
            "sections": sections,
            "log_widget": log_widget,
        }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the create_child_workspace workflow."""

from typing import Any

import debusine.worker.tags as wtags
from debusine.db.models import TaskDatabase, WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.server.workflows import CreateChildWorkspaceWorkflow
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import ChildWorkspaceData
from debusine.tasks.models import BaseDynamicTaskData
from debusine.test.django import TestCase


class CreateChildWorkspaceWorkflowTests(TestCase):
    """Unit tests for :py:class:`CreateChildWorkspaceWorkflow`."""

    scenario = scenarios.DefaultContext()

    def create_workflow(
        self, task_data: dict[str, Any]
    ) -> CreateChildWorkspaceWorkflow:
        """Create a child workspace workflow."""
        wr = self.playground.create_workflow(
            task_name="create_child_workspace", task_data=task_data
        )
        workflow = wr.get_task()
        assert isinstance(workflow, CreateChildWorkspaceWorkflow)
        return workflow

    def test_compute_system_required_tags(self) -> None:
        workflow = self.create_workflow({"suffix": "test"})
        self.assertCountEqual(
            workflow.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    def test_compute_dynamic_data(self) -> None:
        workflow = self.create_workflow({"suffix": "test"})

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(workflow.work_request)),
            BaseDynamicTaskData(subject="test", parameter_summary="test"),
        )

    def test_get_label_without_parameter_summary(self) -> None:
        """Test get_label."""
        workflow = self.create_workflow({"suffix": "test_1"})
        self.assertEqual(
            workflow.get_label(), "create_child_workspace-template"
        )

    def test_get_label_with_parameter_summary(self) -> None:
        """Test get_label."""
        root = self.orchestrate(task_data=ChildWorkspaceData(suffix="test_2"))

        self.assertEqual(
            root.get_label(), "create_child_workspace-template: test_2"
        )

    def test_create_orchestrator(self) -> None:
        """Workflow can be instantiated."""
        workflow = self.create_workflow({"suffix": "test"})
        self.assertIsNone(workflow.data.prefix)
        self.assertEqual(workflow.data.suffix, "test")
        self.assertTrue(workflow.data.public)
        self.assertIsNone(workflow.data.owner_group)
        self.assertEqual(workflow.data.workflow_template_names, [])
        self.assertEqual(workflow.data.expiration_delay, 60)

    def orchestrate(
        self,
        task_data: ChildWorkspaceData,
        *,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create and orchestrate a CreateChildWorkspaceWorkflow."""
        template = self.playground.create_workflow_template(
            name="create_child_workspace-template",
            task_name="create_child_workspace",
            workspace=workspace,
        )
        root = self.playground.create_workflow(
            task_name="create_child_workspace",
            task_data=task_data.model_dump(mode="json"),
        )
        self.assertEqual(root.workspace, template.workspace)
        self.assertTrue(orchestrate_workflow(root))
        return root

    def test_populate(self) -> None:
        root = self.orchestrate(task_data=ChildWorkspaceData(suffix="test"))
        # There is only one child task
        create_child_workspace = root.children.get()
        self.assertEqual(
            create_child_workspace.task_name, "create_child_workspace"
        )
        self.assertEqual(
            create_child_workspace.task_data,
            {
                "prefix": root.workspace.name,
                "suffix": "test",
                "expiration_delay": 60,
                "owner_group": None,
                "public": True,
                "workflow_template_names": [],
            },
        )

    def test_populate_prefix(self) -> None:
        root = self.orchestrate(
            task_data=ChildWorkspaceData(prefix="namespace", suffix="test")
        )
        # There is only one child task
        create_child_workspace = root.children.get()
        self.assertEqual(
            create_child_workspace.task_name, "create_child_workspace"
        )
        self.assertEqual(
            create_child_workspace.task_data,
            {
                "prefix": "namespace",
                "suffix": "test",
                "expiration_delay": 60,
                "owner_group": None,
                "public": True,
                "workflow_template_names": [],
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Wait task to delay until a given time is reached."""

from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.db.models.tasks import DefaultDynamicData
from debusine.server.tasks.wait import BaseWaitTask
from debusine.server.tasks.wait.models import DelayData
from debusine.tasks.models import BaseDynamicTaskData


class Delay(
    BaseWaitTask[DelayData, BaseDynamicTaskData], DefaultDynamicData[DelayData]
):
    """Task that delays until a given time is reached."""

    TASK_VERSION = 1

    @override
    def _execute(self) -> WorkRequestResults:
        """Never called; the scheduler completes these tasks directly."""
        raise NotImplementedError()

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return f"delay until {self.data.delay_until}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the autopkgtest workflow."""

import datetime as dt
from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, TypeAlias

from django.test import override_settings
from django.utils import timezone

import debusine.worker.tags as wtags
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestResultStatus,
    DebianAutopkgtestSource,
    DebusinePromise,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    TaskDatabase,
    WorkRequest,
)
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.scheduler import schedule
from debusine.server.workflows import (
    AutopkgtestWorkflow,
    SbuildWorkflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    AutopkgtestWorkflowDynamicData,
    BaseWorkflowData,
    SbuildWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestBase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    AutopkgtestNeedsInternet,
    BackendType,
    BaseDynamicTaskData,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class AutopkgtestWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestBase[AutopkgtestWorkflow]
):
    """Unit tests for :py:class:`AutopkgtestWorkflow`."""

    source_artifact: ClassVar[Artifact]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common data."""
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello"
        )

    def create_autopkgtest_workflow(
        self,
        extra_task_data: dict[str, Any] | None = None,
        parent: WorkRequest | None = None,
        validate: bool = True,
    ) -> AutopkgtestWorkflow:
        """Create an autopkgtest workflow."""
        task_data = {
            "source_artifact": self.source_artifact.pk,
            "binary_artifacts": ["internal@collections/name:build-amd64"],
            "vendor": "debian",
            "codename": "sid",
        }
        if extra_task_data is not None:
            task_data.update(extra_task_data)
        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=task_data,
            parent=parent,
            validate=validate,
        )
        return self.get_workflow(wr)

    def create_binary_upload(
        self, architecture: str, filenames: list[str]
    ) -> Artifact:
        """Create a minimal `debian:upload` artifact with binaries."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data={
                "type": "dpkg",
                "changes_fields": {
                    "Architecture": architecture,
                    "Files": [{"name": filename} for filename in filenames],
                },
            },
        )
        return artifact

    def add_uploads(
        self, work_request: WorkRequest, architectures: Iterable[str]
    ) -> None:
        """Add multiple uploads to a workflow's internal collection."""
        internal_collection = work_request.internal_collection
        assert internal_collection is not None
        for architecture in architectures:
            upload = self.create_binary_upload(
                architecture, [f"hello_1.0-1_{architecture}.deb"]
            )
            internal_collection.manager.add_artifact(
                upload,
                user=work_request.created_by,
                workflow=work_request,
                name=f"build-{architecture}",
                variables={"architecture": architecture},
            )

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        *,
        timestamp: dt.datetime | None = None,
    ) -> CollectionItem:
        """Add an autopkgtest result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="autopkgtest",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int((timestamp or timezone.now()).timestamp()),
                "work_request_id": work_request.id,
            },
        )

    def orchestrate(
        self,
        task_data: AutopkgtestWorkflowData,
        architectures: Sequence[str],
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate an AutopkgtestWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow that runs sbuild and autopkgtest."""

            TASK_NAME = pipeline_task_name

            def populate(self_) -> None:
                """Populate the pipeline."""
                sbuild = self_.work_request.create_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(
                            source_artifact=task_data.source_artifact
                        ),
                        target_distribution="debian:sid",
                        architectures=["all", *architectures],
                    ),
                )
                sbuild_workflow = sbuild.get_task()
                assert isinstance(sbuild_workflow, SbuildWorkflow)
                source_artifact = (
                    workflow_utils.locate_debian_source_package_lookup(
                        sbuild_workflow,
                        "input.source_artifact",
                        task_data.source_artifact,
                    )
                )
                self.playground.advance_work_request(sbuild, mark_running=True)
                for architecture in ("all", *architectures):
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(source_artifact=source_artifact),
                            build_architecture=architecture,
                            environment="debian/match:codename=sid",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.UPLOAD,
                        f"build-{architecture}",
                        data={"architecture": architecture},
                    )
                sbuild.unblock_workflow_children()

                autopkgtest = self_.work_request.create_child_workflow(
                    task_name="autopkgtest",
                    task_data=task_data,
                )
                self.playground.advance_work_request(
                    autopkgtest, mark_pending=True
                )
                self_.orchestrate_child(autopkgtest)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(orchestrate_workflow(root))

        return root

    def test_create_orchestrator(self) -> None:
        """An AutopkgtestWorkflow can be instantiated."""
        source_artifact = self.source_artifact.pk
        binary_artifacts = ("internal@collections/name:build-arm64",)
        vendor = "debian"
        codename = "trixie"
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "source_artifact": source_artifact,
                "binary_artifacts": binary_artifacts,
                "vendor": vendor,
                "codename": codename,
            },
            validate=False,
        )

        self.assertEqual(w.data.source_artifact, source_artifact)
        self.assertEqual(
            w.data.binary_artifacts, LookupMultiple(binary_artifacts)
        )
        self.assertEqual(w.data.vendor, vendor)
        self.assertEqual(w.data.codename, codename)
        self.assertEqual(w.data.backend, BackendType.UNSHARE)

    def test_create_orchestrator_explicit_backend(self) -> None:
        """An AutopkgtestWorkflow can be instantiated with a backend."""
        w = self.create_autopkgtest_workflow(
            extra_task_data={"backend": BackendType.INCUS_LXC}, validate=False
        )

        self.assertEqual(w.data.backend, BackendType.INCUS_LXC)

    def test_architectures(self) -> None:
        """Workflow uses architectures from binary_artifacts."""
        architectures = ("amd64", "arm64", "i386")
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "binary_artifacts": [
                    f"internal@collections/name:build-{architecture}"
                    for architecture in architectures
                ]
            },
            validate=False,
        )
        self.add_uploads(w.work_request, architectures)

        self.assertCountEqual(w.architectures, architectures)

    def test_architectures_arch_indep_and_arch_dep(self) -> None:
        """Workflow handles arch-indep plus arch-dep binary_artifacts."""
        architectures = ("all", "arm64", "armhf")
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "binary_artifacts": [
                    f"internal@collections/name:build-{architecture}"
                    for architecture in architectures
                ]
            },
            validate=False,
        )
        self.add_uploads(w.work_request, architectures)

        self.assertEqual(w.architectures, {"arm64", "armhf"})

    def test_architectures_arch_indep_only(self) -> None:
        """Workflow handles only having arch-indep binary_artifacts."""
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "binary_artifacts": ["internal@collections/name:build-all"]
            },
            validate=False,
        )
        self.add_uploads(w.work_request, ["all"])

        self.assertEqual(w.architectures, {"amd64"})

    def test_architectures_arch_indep_only_arch_all_build_architecture(
        self,
    ) -> None:
        """Workflow honours `arch_all_build_architecture`."""
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "binary_artifacts": ["internal@collections/name:build-all"],
                "arch_all_build_architecture": "s390x",
            },
            validate=False,
        )
        self.add_uploads(w.work_request, ["all"])

        self.assertEqual(w.architectures, {"s390x"})

    def test_architectures_intersect_task_data(self) -> None:
        """Setting architectures in task data constrains the set."""
        architectures = ("amd64", "arm64", "i386")
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "binary_artifacts": [
                    f"internal@collections/name:build-{architecture}"
                    for architecture in architectures
                ],
                "architectures": ["amd64", "i386"],
            },
            validate=False,
        )
        self.add_uploads(w.work_request, architectures)

        self.assertEqual(w.architectures, {"amd64", "i386"})

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        w = self.create_autopkgtest_workflow(validate=False)
        self.add_uploads(w.work_request, ("amd64",))

        w.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        w = self.create_autopkgtest_workflow(
            extra_task_data={"qa_suite": "nonexistent@debian:suite"},
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            w.validate_input()

    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        w = self.create_autopkgtest_workflow(
            extra_task_data={
                "reference_qa_results": "nonexistent@debian:qa-results"
            },
            validate=False,
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            w.validate_input()

    def test_validate_input_architecture_errors(self) -> None:
        """validate_input raises errors in computing architectures."""
        w = self.create_autopkgtest_workflow(validate=False)
        assert w.work_request.internal_collection is not None
        # Create a promise with no architecture.
        self.playground.create_bare_data_item(
            w.work_request.internal_collection,
            "build-amd64",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=w.work_request.id + 1,
                promise_workflow_id=w.work_request.id,
                promise_category=ArtifactCategory.UPLOAD,
            ),
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "Cannot determine architecture for lookup result",
        ):
            w.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "other", "1.0-1", "amd64")
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def test_has_current_reference_qa_result_too_old(self) -> None:
        """_has_current_reference_qa_result: result is too old."""
        now = timezone.now()
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            timestamp=now - dt.timedelta(days=31),
        )
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        source_item = sid_suite.manager.add_artifact(
            source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        source_item.created_at = now
        source_item.save()

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple(()),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result("amd64")
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """The workflow populates child work requests."""
        architectures = ("amd64", "i386")
        source_artifact = self.playground.create_source_artifact()

        root = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", *architectures)
                    ]
                ),
                vendor="debian",
                codename="sid",
                extra_repositories=[
                    ExtraExternalRepository(
                        url="http://example.com/",
                        suite="bookworm",
                        components=["main"],
                    )
                ],
            ),
            architectures=architectures,
        )

        autopkgtest = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest", parent=root
        )
        children = list(
            WorkRequest.objects.filter(parent=autopkgtest).order_by("id")
        )
        self.assertEqual(len(children), len(architectures))
        for child, architecture in zip(children, architectures):
            self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
            self.assertEqual(child.task_type, TaskTypes.WORKER)
            self.assertEqual(child.task_name, "autopkgtest")
            self.assertEqual(
                child.task_data,
                {
                    "input": {
                        "source_artifact": autopkgtest.task_data[
                            "source_artifact"
                        ],
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/"
                                f"name:build-{architecture}",
                                "internal@collections/name:build-all",
                            ]
                        ),
                        "context_artifacts": [],
                    },
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=sid",
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                    "backend": BackendType.UNSHARE,
                    "include_tests": [],
                    "exclude_tests": [],
                    "debug_level": 0,
                    "extra_environment": {},
                    "needs_internet": AutopkgtestNeedsInternet.RUN,
                    "fail_on": {},
                    "timeout": None,
                },
            )
            self.assert_work_request_event_reactions(
                child,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        collection="internal@collections",
                        name_template=f"autopkgtest-{architecture}",
                        artifact_filters={
                            "category": ArtifactCategory.AUTOPKGTEST
                        },
                    )
                ],
            )
            self.assertQuerySetEqual(
                child.dependencies.all(),
                list(
                    WorkRequest.objects.filter(
                        task_type=TaskTypes.WORKER,
                        task_name="sbuild",
                        task_data__build_architecture__in={architecture, "all"},
                    )
                ),
            )
            self.assertEqual(
                child.workflow_data_json,
                {
                    "display_name": f"autopkgtest {architecture}",
                    "step": f"autopkgtest-{architecture}",
                },
            )

        # Population is idempotent.
        self.get_workflow(autopkgtest).populate()
        children = list(WorkRequest.objects.filter(parent=autopkgtest))
        self.assertEqual(len(children), len(architectures))

    @preserve_db_task_registry()
    def test_populate_upload(self) -> None:
        """The workflow accepts debian:upload source artifacts."""
        architectures = ("amd64", "i386")
        upload_artifacts = self.playground.create_upload_artifacts()

        root = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", *architectures)
                    ]
                ),
                vendor="debian",
                codename="sid",
                extra_repositories=[
                    ExtraExternalRepository(
                        url="http://example.com/",
                        suite="bookworm",
                        components=["main"],
                    )
                ],
            ),
            architectures=architectures,
        )

        autopkgtest = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="autopkgtest",
            parent=root,
        )
        children = list(
            WorkRequest.objects.filter(parent=autopkgtest).order_by("id")
        )
        self.assertEqual(len(children), len(architectures))
        for child, architecture in zip(children, architectures):
            self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
            self.assertEqual(child.task_type, TaskTypes.WORKER)
            self.assertEqual(child.task_name, "autopkgtest")
            self.assertEqual(
                child.task_data,
                {
                    "input": {
                        "source_artifact": (
                            f"{upload_artifacts.source.id}@artifacts"
                        ),
                        "binary_artifacts": sorted(
                            [
                                f"internal@collections/"
                                f"name:build-{architecture}",
                                "internal@collections/name:build-all",
                            ]
                        ),
                        "context_artifacts": [],
                    },
                    "build_architecture": architecture,
                    "environment": "debian/match:codename=sid",
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                    "backend": BackendType.UNSHARE,
                    "include_tests": [],
                    "exclude_tests": [],
                    "debug_level": 0,
                    "extra_environment": {},
                    "needs_internet": AutopkgtestNeedsInternet.RUN,
                    "fail_on": {},
                    "timeout": None,
                },
            )
            self.assert_work_request_event_reactions(
                child,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        collection="internal@collections",
                        name_template=f"autopkgtest-{architecture}",
                        artifact_filters={
                            "category": ArtifactCategory.AUTOPKGTEST
                        },
                    )
                ],
            )
            self.assertQuerySetEqual(
                child.dependencies.all(),
                list(
                    WorkRequest.objects.filter(
                        task_type=TaskTypes.WORKER,
                        task_name="sbuild",
                        task_data__build_architecture__in={architecture, "all"},
                    )
                ),
            )
            self.assertEqual(
                child.workflow_data_json,
                {
                    "display_name": f"autopkgtest {architecture}",
                    "step": f"autopkgtest-{architecture}",
                },
            )

        # Population is idempotent.
        self.get_workflow(autopkgtest).populate()
        children = list(WorkRequest.objects.filter(parent=autopkgtest))
        self.assertEqual(len(children), len(architectures))

    def test_populate_mixed_uploads(self) -> None:
        """The workflow accepts mixed debian:upload artifacts in all inputs."""
        artifacts = self.playground.create_upload_artifacts(
            binaries=[
                ("hello", "amd64"),
                ("hello", "i386"),
                ("hello-doc", "all"),
            ]
        )
        context = self.playground.create_upload_artifacts(
            binaries=[
                ("context", "amd64"),
                ("context", "i386"),
                ("context-common", "all"),
            ]
        )
        architectures = ["amd64", "i386"]

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                source_artifact=artifacts.upload.id,
                binary_artifacts=LookupMultiple((artifacts.upload.id,)),
                context_artifacts=LookupMultiple((context.upload.id,)),
                architectures=architectures,
                vendor="debian",
                codename="sid",
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.order_by("id").all())
        self.assertEqual(len(children), len(architectures))
        for architecture, child in zip(architectures, children):
            with self.subTest(architecture=architecture):
                self.assertEqual(
                    child.task_data["build_architecture"], architecture
                )
                self.assertEqual(
                    child.task_data["input"],
                    {
                        "source_artifact": (f"{artifacts.source.id}@artifacts"),
                        "binary_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in artifacts.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                        "context_artifacts": sorted(
                            f"{artifact.id}@artifacts"
                            for artifact in context.binaries
                            if artifact.data["deb_fields"]["Architecture"]
                            in (architecture, "all")
                        ),
                    },
                )

    def test_populate_experimental(self) -> None:
        """The workflow handles overlay distributions."""
        source_artifact = self.playground.create_source_artifact()
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        self.playground.create_debian_environment(
            codename="experimental", variant="autopkgtest"
        )
        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="experimental",
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(len(child.task_data["extra_repositories"]), 1)
        repo = child.task_data["extra_repositories"][0]
        self.assertEqual(repo["suite"], "experimental")

    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")

        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), [])

    def test_populate_previous_reference_qa_result_too_old(self) -> None:
        """The workflow produces a reference QA result if it is outdated."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        old_qa_result = self.add_qa_result(
            sid_qa_results,
            "hello",
            "1.0-1",
            "amd64",
            timestamp=timezone.now() - dt.timedelta(days=31),
        )

        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        source_item = sid_suite.manager.add_artifact(
            source_artifact,
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )
        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.get_workflow(wr).populate()

        [child] = wr.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "autopkgtest")
        self.assertEqual(
            child.task_data,
            {
                "input": {
                    "source_artifact": source_artifact.id,
                    "binary_artifacts": [f"{binary_artifact.id}@artifacts"],
                    "context_artifacts": [],
                },
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "extra_repositories": None,
                "backend": BackendType.UNSHARE,
                "include_tests": [],
                "exclude_tests": [],
                "debug_level": 0,
                "extra_environment": {},
                "needs_internet": AutopkgtestNeedsInternet.RUN,
                "fail_on": {},
                "timeout": None,
            },
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-1",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
            artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[
                ActionSkipIfLookupResultChanged(
                    lookup=(
                        f"sid@{CollectionCategory.QA_RESULTS}/"
                        f"latest:autopkgtest_hello_amd64"
                    ),
                    collection_item_id=old_qa_result.id,
                    promise_name="reference-qa-result|autopkgtest-amd64",
                )
            ],
            on_failure=[qa_result_action],
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    collection="internal@collections",
                    name_template="reference-qa-result|autopkgtest-amd64",
                    artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
                ),
                qa_result_action,
            ],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.AUTOPKGTEST, work_request=child
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:autopkgtest_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.playground.create_debian_environment(
            codename="sid", variant="autopkgtest"
        )

        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                architecture="amd64",
            )
        )
        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="sid",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
            ),
        )

        self.get_workflow(wr).populate()

        racing_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        [child] = wr.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'sid@{CollectionCategory.QA_RESULTS}/"
                    f"latest:autopkgtest_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/"
                f"latest:autopkgtest_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = [
            "amd64",
            "arm64",
            "i386",
            "ppc64el",
            "riscv64",
            "s390x",
        ]
        reference_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        new_source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-2"
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=True,
                architectures=architectures,
                source_artifact=reference_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                        if arch != "ppc64el"
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=architectures,
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=AutopkgtestWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                architectures=architectures,
                source_artifact=new_source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            architectures=architectures,
            parent=root,
        )

        # Unblock the autopkgtest tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS
            )

        DARS: TypeAlias = DebianAutopkgtestResultStatus

        reference_autopkgtest_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        reference_source_url = (
            f"http://example.com/{reference_source_artifact.get_absolute_url()}"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_autopkgtest_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="autopkgtest"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST,
                data=DebianAutopkgtest(
                    results={
                        name: DebianAutopkgtestResult(status=status)
                        for name, status in statuses.items()
                    },
                    cmdline="unused",
                    source_package=DebianAutopkgtestSource(
                        name="hello", version="1.0-1", url=reference_source_url
                    ),
                    architecture=architecture,
                    distribution="sid",
                ),
                workspace=reference.workspace,
                work_request=reference_tasks[architecture],
            )[0]
            for architecture, statuses in (
                ("amd64", {"upstream": DARS.PASS, "old": DARS.PASS}),
                ("arm64", {"upstream": DARS.PASS}),
                ("i386", {"upstream": DARS.FAIL}),
                ("riscv64", {"upstream": DARS.FAIL, "performance": DARS.FAIL}),
                ("s390x", {"upstream": DARS.FAIL}),
            )
        }
        new_autopkgtest_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="autopkgtest"
        )
        new_source_url = (
            f"http://example.com/{new_source_artifact.get_absolute_url()}"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_autopkgtest_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="autopkgtest"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.AUTOPKGTEST,
                data=DebianAutopkgtest(
                    results={
                        name: DebianAutopkgtestResult(status=status)
                        for name, status in statuses.items()
                    },
                    cmdline="unused",
                    source_package=DebianAutopkgtestSource(
                        name="hello", version="1.0-2", url=new_source_url
                    ),
                    architecture=architecture,
                    distribution="sid",
                ),
                workspace=new.workspace,
                work_request=new_tasks[architecture],
            )[0]
            for architecture, statuses in (
                ("amd64", {"upstream": DARS.PASS}),
                ("arm64", {"upstream": DARS.FAIL}),
                ("i386", {"upstream": DARS.PASS}),
                ("ppc64el", {"new": DARS.PASS}),
                ("riscv64", {"upstream": DARS.FAIL, "performance": DARS.PASS}),
                ("s390x", {"upstream": DARS.PASS, "new": DARS.FAIL}),
            )
        }

        self.assertIsNone(new_autopkgtest_workflow.output_data)
        self.assertEqual(schedule(), [])
        self.playground.advance_work_request(
            reference_tasks["amd64"], result=WorkRequest.Results.SUCCESS
        )
        self.assertEqual(schedule(), [])
        self.playground.advance_work_request(
            new_tasks["amd64"], result=WorkRequest.Results.SUCCESS
        )

        self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)

        expected_analysis: list[
            tuple[str, RegressionAnalysisStatus, dict[str, str]]
        ] = [
            (
                "amd64",
                RegressionAnalysisStatus.STABLE,
                {"upstream": "stable", "old": "stable"},
            )
        ]
        new_autopkgtest_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_autopkgtest_workflow,
            "1.0-1",
            reference_results,
            "1.0-2",
            new_results,
            expected_analysis,
        )

        expected_callback_statuses = {
            "amd64": WorkRequest.Statuses.COMPLETED,
            "arm64": WorkRequest.Statuses.BLOCKED,
            "i386": WorkRequest.Statuses.BLOCKED,
            "ppc64el": WorkRequest.Statuses.BLOCKED,
            "riscv64": WorkRequest.Statuses.BLOCKED,
            "s390x": WorkRequest.Statuses.BLOCKED,
        }
        self.assert_callback_statuses(
            new_autopkgtest_workflow, expected_callback_statuses
        )

        for architecture, reference_result, new_result, status, details in (
            (
                "arm64",
                WorkRequest.Results.SUCCESS,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
                {"upstream": "regression"},
            ),
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
                {"upstream": "improvement"},
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
                {"new": "stable"},
            ),
            (
                "riscv64",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.IMPROVEMENT,
                {"upstream": "stable", "performance": "improvement"},
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
                {"upstream": "improvement", "new": "regression"},
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture], result=reference_result
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result
            )
            expected_analysis.append((architecture, status, details))
            expected_callback_statuses[architecture] = (
                WorkRequest.Statuses.PENDING
            )

        self.assert_callback_statuses(
            new_autopkgtest_workflow, expected_callback_statuses
        )

        # Running any one of the pending callbacks is enough, but in fact
        # one is scheduled for each architecture and the scheduler will
        # catch up with them over multiple runs.  Simulate this and check
        # that the result is correct at each stage.
        for architecture in ("arm64", "i386", "ppc64el", "riscv64", "s390x"):
            self.schedule_and_run_workflow_callback(new_autopkgtest_workflow)

            new_autopkgtest_workflow.refresh_from_db()
            self.assert_regression_analysis(
                new_autopkgtest_workflow,
                "1.0-1",
                reference_results,
                "1.0-2",
                new_results,
                expected_analysis,
            )
            expected_callback_statuses[architecture] = (
                WorkRequest.Statuses.COMPLETED
            )
            self.assert_callback_statuses(
                new_autopkgtest_workflow, expected_callback_statuses
            )

    def test_compute_system_required_tags(self) -> None:
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
                context_artifacts=LookupMultiple((context_artifact.id,)),
            ),
        )
        workflow = self.get_workflow(wr)
        self.assertCountEqual(
            workflow.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    def test_compute_dynamic_data(self) -> None:
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
                context_artifacts=LookupMultiple((context_artifact.id,)),
            ),
        )
        workflow = self.get_workflow(wr)

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(wr)),
            AutopkgtestWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=source_artifact.id,
                binary_artifacts_ids=[binary_artifact.id],
                context_artifacts_ids=[context_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )

        wr = self.playground.create_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple((binary_artifact.id,)),
                vendor="debian",
                codename="trixie",
                context_artifacts=LookupMultiple((context_artifact.id,)),
            ),
        )
        workflow = self.get_workflow(wr)

        workflow.dynamic_data = workflow.compute_dynamic_data(TaskDatabase(wr))

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=source_artifact.id,
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[binary_artifact.id],
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.context_artifacts,
                    label="context_artifacts",
                    artifact_ids=[context_artifact.id],
                ),
            ],
        )
